/****************************************************************************
*
*  $RCSfile: siconv.c,v $
*
*   Converts a stream from one charset to another.
*  Original Author: Caskey Dickson <caskey@technocage.com> 2001-02-28
*  Copyright 2001 TechnoCage, Inc.  All Rights Reserved
*  $Id: siconv.c,v 1.5 2001/03/04 03:53:36 caskey Exp $
*
*  This code based upon web copy of HP-UX man page found online via
*  google.  Linux needs an iconv man page already.  Sheesh.
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software Foundation, 
*   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*****************************************************************************/


#include <iconv.h>
#include <errno.h>
#include <stdio.h>     /* printf() */
#include <unistd.h>    /* read() & write() */
#include <string.h>    /* memmove() */

#define FATAL "FATAL"

#define BAD_OPEN "Bad open (most likely unsupported char set specfied)"
#define BAD_CLOSE "Bad close (I don't know what this is, please email <caskey@technocage.com>"
#define BAD_CONVERSION "Bad conversion (illegal character in input data)"

#define GOOD 0
#define BAD 1

int bufferSize = BUFSIZ;

void error(char* type, char* code){
  fprintf(stderr, "Error %s, %s\n", type, code);
  exit(1);
}

int convert(fromcode, tocode, input, output)
char *tocode;                       /* tocode name */
char *fromcode;                     /* fromcode name */
int input;                          /* input file descriptor */
int output;                          /* output file descriptor */
{
  iconv_t cd;                     /* conversion descriptor */

  int bytesread;                  /* num bytes read into input buffer */
  unsigned char inbuf[BUFSIZ];    /* input buffer */
  unsigned char *inchar;          /* ptr to input character */
  int inbytesleft;                /* num bytes left in input buffer */
  unsigned char outbuf[BUFSIZ];   /* output buffer */
  unsigned char *outchar;         /* ptr to output character */
  int outbytesleft;               /* num bytes left in output buffer */
  size_t ret_val;                 /* number of conversions */

  /* Initiate conversion -- get conversion descriptor */
  if ((cd = iconv_open(tocode, fromcode)) == (iconv_t)-1) {
    error(FATAL, BAD_OPEN);
  }

  inbytesleft = 0;                        /* no. of bytes converted */
  /* translate the characters */
  for ( ;; ) {
    /*
     * if any bytes are leftover, they will be in the
     * beginning of the buffer on the next read().
     */

    inchar = inbuf;          /* points to input buffer */
    outchar = outbuf;        /* points to output buffer */
    outbytesleft = bufferSize;   /* no of bytes to be converted */
    if ((bytesread = read(input, inbuf+inbytesleft,
        (size_t)bufferSize-inbytesleft)) < 0) {
      perror("read error");
      return BAD;
    }
    if (!(inbytesleft += bytesread)) {
      break;           /* end of conversions */
    }

    ret_val = iconv(cd, (const char**) &inchar, &inbytesleft,
        (char**) &outchar, &outbytesleft);

    if (write(output, outbuf, (size_t)bufferSize-outbytesleft) < 0) {
      perror("prog2");
      return BAD;
    }

    /* iconv() returns the number of non-identical conversions
     * performed. If the entire string in the input buffer is
     * converted, the value pointed to by inbytesleft will be
     * zero. If the conversion stopped due to any reason, the
     * value pointed to by inbytesleft will be non-zero and
     * errno is set to indicate the condition.
     */
    if ((ret_val == -1) && (errno == EINVAL)) {

      /* Input conversion stopped due to an incomplete
       * character or shift sequence at the end of the
       * input buffer.
       */

      /* Copy data left, to the start of buffer */
      memmove((char *)inbuf, (char *)inchar, (size_t)inbytesleft);
    } else if ((ret_val == -1) && (errno == EILSEQ)) {
      /* Input conversion stopped due to an input byte
       * that does not belong to the input codeset.
       */
      error(FATAL, BAD_CONVERSION);
    } else if ((ret_val == -1) && (errno == E2BIG)) {
      /* Input conversion stopped due to lack of space
       * in the output buffer. inbytesleft has the
       * number of bytes to be converted.
       */
      memmove((char *)inbuf, (char *)inchar, (size_t)inbytesleft);
    }
    /* Go back and read from the input file. */
  }
  /* end conversion & get rid of the conversion table */
  if (iconv_close(cd) == BAD) {
    error(FATAL, BAD_CLOSE);
  }
  return GOOD;
}

void usage(void) {
  printf("usage:\n");
  printf("\n");
  printf("  convert stdin from iso-8859-1 to utf8 on stdout\n");
  printf("  $ siconv\n");
  printf("  convert stdin from charset-A to charset-B on stdout\n");
  printf("  $ siconv charset-A charset-B\n");
  printf("\n");
  exit(1);
}

int main(int argc, char *argv[]) {
  char *from = "iso-8859-1";
  char *to = "unicode";

  if ( argc == 3 ) {
    from = argv[1];
    to = argv[2];
  } else if ( argc != 1 ) {
    usage();
  }
  fprintf(stderr, "Converting from %s to %s\n", from, to);
  convert(from, to, STDIN_FILENO, STDOUT_FILENO);

  return 0;
}

